use anyhow::{Context, Result};

use cpal::traits::{DeviceTrait, StreamTrait};
use cpal::{FromSample, Sample, SampleFormat};

use tracing::{info, trace, warn};

use std::collections::VecDeque;

pub mod foreign_signal;
use foreign_signal::ForeignTwoStateCarrier;

mod input_dsp;
use self::input_dsp::take_input_pcm;

pub fn audio_receiver_main(
    tx_for_demod: tokio::sync::mpsc::Sender<Vec<f64>>,
    cpal_input: &cpal::Device,
) -> Result<cpal::Stream> {
    trace!("audio_receiver started");
    let cpal_input_config = cpal_input.supported_input_configs()
        .context("could not find a suitable audio input configuration. are you root, trying to use a normal user's audio stack? try setting the capability CAP_NET_ADMIN=ep on this binary, and running as the user who is running the audio stack.")?
        .max_by_key(|cfg| {
            if cfg.channels() > 1 { return -999; }
            match cfg.sample_format() {
                SampleFormat::U8 => 10,
                SampleFormat::I8 => 11,
                SampleFormat::U16 => 20,
                SampleFormat::I16 => 21,
                SampleFormat::U32 => 30,
                SampleFormat::I32 => 31,
                SampleFormat::F32 => 40,
                SampleFormat::F64 => 50,
                SampleFormat::U64 => 60,
                SampleFormat::I64 => 61,
                x => { warn!("Backend offered unknown sample format {x}",); -999 }
            }
    }).expect("input device has no supported configurations available")
        .with_max_sample_rate();
        //.with_sample_rate(cpal::SampleRate(8000));

    let sample_rate = cpal_input_config.sample_rate();

    info!(
        "input samples are {} @ {}hz", 
        &cpal_input_config.sample_format(),
        sample_rate.0,
    );

    let failed_pcm_take = move |e| {
        warn!("dropped some PCM data on input stream?! {}", e);
    };

    let mut pcm_buf: VecDeque<rustfft::num_complex::Complex<f64>> =
        vec![].into();
    // TODO: determine freqs from handshake
    let mut foreign_carrier = ForeignTwoStateCarrier {
        acquired: false,
        clock_hz: 1200.0,
        freq_low: Some(2400.0),
        freq_high: Some(3600.0),
    };

    let input_stream = match cpal_input_config.sample_format() {
        cpal::SampleFormat::U8 => cpal_input.build_input_stream(
            &cpal_input_config.into(),
            move |pcm, _: &_| {
                take_input_pcm::<u8>(pcm, sample_rate, &tx_for_demod, &mut pcm_buf, &mut foreign_carrier)
            },
            failed_pcm_take,
            None,
        ).context("failed to build input stream")?,
        cpal::SampleFormat::I8 => cpal_input.build_input_stream(
            &cpal_input_config.into(),
            move |pcm, _: &_| {
                take_input_pcm::<i8>(pcm, sample_rate, &tx_for_demod, &mut pcm_buf, &mut foreign_carrier)
            },
            failed_pcm_take,
            None,
        ).context("failed to build input stream")?,
        cpal::SampleFormat::I16 => cpal_input.build_input_stream(
            &cpal_input_config.into(),
            move |pcm, _: &_| {
                take_input_pcm::<i16>(pcm, sample_rate, &tx_for_demod, &mut pcm_buf, &mut foreign_carrier)
            },
            failed_pcm_take,
            None,
        ).context("failed to build input stream")?,
        cpal::SampleFormat::I32 => cpal_input.build_input_stream(
            &cpal_input_config.into(),
            move |pcm, _: &_| {
                take_input_pcm::<i32>(pcm, sample_rate, &tx_for_demod, &mut pcm_buf, &mut foreign_carrier)
            },
            failed_pcm_take,
            None,
        ).context("failed to build input stream")?,
        cpal::SampleFormat::F32 => cpal_input.build_input_stream(
            &cpal_input_config.into(),
            move |pcm, _: &_| {
                take_input_pcm::<f32>(pcm, sample_rate, &tx_for_demod, &mut pcm_buf, &mut foreign_carrier)
            },
            failed_pcm_take,
            None,
        ).context("failed to build input stream")?,
        cpal::SampleFormat::F64 => cpal_input.build_input_stream(
            &cpal_input_config.into(),
            move |pcm, _: &_| {
                take_input_pcm::<f64>(pcm, sample_rate, &tx_for_demod, &mut pcm_buf, &mut foreign_carrier)
            },
            failed_pcm_take,
            None,
        ).context("failed to build input stream")?,
        cpal::SampleFormat::U64 => cpal_input.build_input_stream(
            &cpal_input_config.into(),
            move |pcm, _: &_| {
                take_input_pcm::<u64>(pcm, sample_rate, &tx_for_demod, &mut pcm_buf, &mut foreign_carrier)
            },
            failed_pcm_take,
            None,
        ).context("failed to build input stream")?,
        cpal::SampleFormat::I64 => cpal_input.build_input_stream(
            &cpal_input_config.into(),
            move |pcm, _: &_| {
                take_input_pcm::<i64>(pcm, sample_rate, &tx_for_demod, &mut pcm_buf, &mut foreign_carrier)
            },
            failed_pcm_take,
            None,
        ).context("failed to build input stream")?,
        sample_format => {
            Err(anyhow::Error::msg(
                format!(
                "unsupported sample format '{}'! please ask developers to add it", sample_format)
            ))?
        }
    };

    input_stream.play().context("failed to start intake audio stream")?;

    Ok(input_stream)
}
